/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

// @(#) internal-src/lib/dacs/cell/common/dacsi_q.h v1.11 - 9/24/07 13:33:05 @(#)

#ifndef _DACSI_Q_H_
#define _DACSI_Q_H_

#include <stdint.h>
#include <assert.h>

#include <dacs.h>
#include <dacs_tags.h>
#include <dacsi_wids.h>

//
//  C++ interface
//
#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

typedef struct dacsi_Qe
{
    uint16_t next;
    uint16_t prev;
    uint32_t off;
} dacsi_Qe_t;

// 
// Queue header structure
// When embedded, this structure must be the first thing so pointers to the
// consuming queue work as generic queue pointers.
typedef struct dacsi_Q
{
    uint32_t   lock;
    dacsi_Qe_t active;
    dacsi_Qe_t free;
} dacsi_Q_t;

#define DACSI_64_TO_32(__addr)  ((unsigned long)(__addr))

#define DACSI_Q_INIT(__Q, __cnt, __etype)                               \
    do {                                                                \
        int i;                                                          \
        __etype *__Qe = (__Q)->entries;                                 \
                                                                        \
        for (i = 0; i < __cnt; i++) {                                   \
            __Qe[i].Q.off =                                             \
                (unsigned int)((unsigned long)&__Qe[i] -                \
                        (unsigned long)(__Q));                          \
            __Qe[i].Q.next = (i < ((__cnt)-1))                          \
                ? (unsigned int)((unsigned long)&__Qe[i+1] -            \
                        (unsigned long)(__Q))                           \
                : (unsigned int)(offsetof(dacsi_Q_t, free));            \
            __Qe[i].Q.prev = (i > 0)                                    \
                ? (unsigned int)((unsigned long)&__Qe[i-1] -            \
                        (unsigned long)(__Q))                           \
                : (unsigned int)(offsetof(dacsi_Q_t, free));            \
        }                                                               \
        (__Q)->Q.free.next = __Qe[0].Q.off;                             \
        (__Q)->Q.free.prev = __Qe[(__cnt)-1].Q.off;                     \
        (__Q)->Q.free.off = offsetof(dacsi_Q_t, free);                  \
        (__Q)->Q.active.next = (__Q)->Q.active.prev =                   \
            offsetof(dacsi_Q_t, active);                                \
        (__Q)->Q.active.off = offsetof(dacsi_Q_t, active);              \
    } while(0)

static inline dacsi_Qe_t *
dacsi_Q_remove(dacsi_Qe_t *Qe)
{
    dacsi_Qe_t *prev = (dacsi_Qe_t *)((unsigned long)Qe - Qe->off + Qe->prev);
    dacsi_Qe_t *next= (dacsi_Qe_t *)((unsigned long)Qe - Qe->off + Qe->next);
    prev->next = Qe->next;
    next->prev = Qe->prev;

    Qe->next = Qe->prev = 0;
    return Qe;
}

static inline dacsi_Qe_t *
dacsi_Q_insert(dacsi_Qe_t *Q, dacsi_Qe_t *Qe)
{
    dacsi_Qe_t *Qprev = (dacsi_Qe_t *)((unsigned long)Q - Q->off + Q->prev);

    // Make sure this is a valid Q entry.
    assert(Qprev->off);

    Qprev->next = Q->prev = Qe->off;
    assert(Qprev->next);
    Qe->prev = Qprev->off;
    assert(Qe->prev);

    Qe->next = Q->off;
    assert(Qe->next);

    return Qe;
}

#define dacsi_Q_getactive(__Q)                                              \
    dacsi_Q_remove((dacsi_Qe_t *)((unsigned long)(__Q) +                    \
                                  ((dacsi_Q_t *)(__Q))->active.next))

#define dacsi_Q_getfree(__Q)                                                \
    dacsi_Q_remove((dacsi_Qe_t *)((unsigned long)(__Q) +                    \
                                  ((dacsi_Q_t *)(__Q))->free.next))

#define dacsi_Q_putactive(__Qe)                                             \
    dacsi_Q_insert(&((dacsi_Q_t *)                                          \
                   ((unsigned long)(__Qe) - (__Qe)->off))->active, (__Qe))

#define dacsi_Q_putfree(__Qe)                                               \
    dacsi_Q_insert(&((dacsi_Q_t *)                                          \
                   ((unsigned long)(__Qe) - (__Qe)->off))->free, (__Qe))

#define dacsi_Q_deactivate(__Qe)                                            \
    dacsi_Q_putfree(dacsi_Q_remove((dacsi_Qe_t *)(__Qe)))

#define dacsi_is_Q_active_empty(__Q)                                        \
    (((dacsi_Q_t *)((void *)__Q))->active.next == ((dacsi_Q_t *)(__Q))->active.off)

#define dacsi_is_Q_free_empty(__Q)                                          \
    (((dacsi_Q_t *)((void *)__Q))->free.next == ((dacsi_Q_t *)(__Q))->free.off)

static inline dacsi_Qe_t *
dacsi_Q_activate(dacsi_Q_t *Q)
{
    dacsi_Qe_t *Qe;
    Qe = dacsi_Q_getfree(Q);
    return dacsi_Q_putactive(Qe);
}



#define    DACSI_JOBQE_CNT  8 

// Signature field for JOB -   "JOBQ"
#define     DACSI_JOB_SIGNATURE   0x4A4C4251

#define DACSI_QE_NEXT(__Qe)                                                 \
    ((dacsi_Qe_t *)((unsigned long)(__Qe) - ((dacsi_Qe_t *)(__Qe))->off +   \
     ((dacsi_Qe_t *)(__Qe))->next))

typedef enum {
#if defined(DACS_PPU)
    DACSI_JOBQ_SEND,
#else
    DACSI_JOBQ_RECV,
#endif
#if defined(DACS_PPU)
    DACSI_JOBQ_RECV,
#else
    DACSI_JOBQ_SEND,
#endif
    DACSI_JOBQ_CNT
} DACSI_JOBQ_TYPE_T;

typedef enum
{
    DACSI_JOB_TYPE_INVALID,    ///< invalid - normally this means the job is 
                               ///            not initialized
    DACSI_JOB_TYPE_SEND,       ///< send - posted by dacs_send
    DACSI_JOB_TYPE_RECV,       ///< recv - posted by dacs_recv
    DACSI_JOB_TYPE_PUT,        ///< DMA put - posted by dacs_put
    DACSI_JOB_TYPE_GET,        ///< DMA get - posted by dacs_get
    DACSI_JOB_TYPE_PUTLIST,    ///< DMA putlist - posted by dacs_put_list
    DACSI_JOB_TYPE_GETLIST,    ///< DMA getlist - posted by dacs_get_list
}   DACSI_JOB_TYPE_T;

/// Status of the job
typedef enum
{
    DACSI_JOB_ST_INVALID,              ///< invalid -  normally this means the 
                                       ///  job is not initialized
    DACSI_JOB_ST_FREE,                 ///< free    -  not being used.
    DACSI_JOB_ST_PENDING,              ///< pending -  data has been Q'ed for 
                                       ///  processing by test/wait
    DACSI_JOB_ST_FINISHED,             ///< finished - data transfer is done.
    DACSI_JOB_ST_COMPLETED,            ///< finished - data transfer is done.
}   DACSI_JOB_ST_T;


typedef struct dacsi_jobQe
{
    dacsi_Qe_t     Q;               //< 8: Queue element data
    uint32_t       stream;          //< 4: Particular stream this job belongs to
    uint32_t       initiator;       //< 4: DE ID of the initiator
    uint32_t       state;           //< 4: Job state (MUST be 16 byte aligned)
    uint32_t       jobid;           //< 4: unique job id
    uint8_t        type;            //< 1: type: invalid, send, recv
    uint8_t        swap;            //< 1: swap byte specified by dacs_recv
    uint8_t        Qtype;           //< 1: Job Q type this job belongs to
    uint8_t        Qidx;            //< 1: 
    unsigned int   error;           //< 4: error status of job
    uint32_t       len;             //< 4: size of buffer
    union {
        uint64_t       ms;          //< 8: buffer specified by dacs_send
        struct {
            uint32_t   unused;
            uint32_t   ls;
        };
    };
    uint8_t         pad[12];
} dacsi_jobQe_t __attribute__ ((aligned(64)));

typedef union dacsi_jobQ
{
    struct
    {
        dacsi_Q_t   Q;
        uint32_t signature;
        dacsi_jobQe_t entries[DACSI_JOBQE_CNT];
    };
    unsigned char   pad[256];
}   dacsi_jobQ_t __attribute__ ((aligned(16)));

extern dacsi_jobQe_t *dacsi_jobQe_peek(dacsi_jobQ_t *, uint32_t, 
                                       DACSI_JOB_TYPE_T);
extern dacsi_jobQe_t *dacsi_jobQe_get(dacsi_jobQ_t *, uint32_t, 
                                      DACSI_JOB_TYPE_T);
extern dacsi_jobQe_t *dacsi_jobQe_alloc(dacsi_jobQ_t *);
extern dacsi_jobQe_t *dacsi_jobQe_free(dacsi_jobQe_t *);
extern dacsi_jobQe_t *dacsi_jobQe_put(dacsi_jobQe_t *);
extern dacsi_jobQe_t *dacsi_create_job(DACSI_JOBQ_TYPE_T, DACS_ERR_T, 
                                       unsigned int, uint32_t, uint64_t, 
                                       uint32_t);
extern void dacsi_jobQe_init(dacsi_jobQe_t *, int, int);
extern void dacsi_jobQ_init(dacsi_jobQ_t *, int, int);
#ifdef NEED_DEBUG
extern void dacsi_jobQ_printjob(dacsi_jobQe_t *);
#endif

extern void dacsi_Q_print_info(dacsi_Q_t *);
extern void dacsi_Q_printQ(dacsi_Q_t *);

typedef struct dacsi_waitQe dacsi_waitQe_t;
struct dacsi_waitQe
{
    dacsi_waitQe_t *next;
    uint64_t job;
    DACSI_JOB_TYPE_T type;
};

typedef struct dacsi_waitQ
{
    dacsi_waitQe_t *first;
    dacsi_waitQe_t *last;
} dacsi_waitQ_t;

DACS_ERR_T dacsi_we_get(dacs_wid_t, de_id_t, dacs_process_id_t, 
                        dacsi_wid_elem_t **, dacsi_waitQe_t **, dacs_tag_t *);
DACS_ERR_T dacsi_waitQe_enq(dacs_wid_t, dacsi_wid_elem_t *, dacsi_waitQe_t *,
                            DACSI_JOB_TYPE_T, dacs_addr_64_t);
DACSI_JOB_TYPE_T dacsi_waitQe_get(dacsi_wid_elem_t *, dacs_addr_64_t *);
DACS_ERR_T dacsi_waitQe_free(dacsi_wid_elem_t *);
DACS_ERR_T dacsi_we_free(dacsi_wait_queue_t *, dacsi_wid_elem_t *);

//
//  C++ interface
//
#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif  //  _DACSI_Q_H_
