/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

// @(#) internal-src/lib/dacs/cell/ppu/dacs_ppe_internal.h v1.37 - 9/24/07 13:33:12 @(#)
//
// PPE-specific internal definitions
//

#ifndef _DACS_PPE_INTERNAL_H_
#define _DACS_PPE_INTERNAL_H_

#include <libspe2.h>
#include <dacs_dps.h>

#include <dacs.h>
#include <dacs_atomic.h>
#include <dacsi_wids.h>
#include <dacs_internal.h>       // base definitions
#include <dacs_ppu_tags.h>

typedef dacs_addr_64_t dacs_context_t;

//  ----    send/recv   ------------------------------------------------
extern dacsi_jobQ_t dacspi_jobQ[1+NUM_SPES_PER_BLADE][DACSI_JOBQ_CNT]; 

extern DACS_ERR_T dacspi_waitQe_enq(dacs_wid_t, dacsi_wid_elem_t *, 
                                    DACSI_JOB_TYPE_T, dacs_addr_64_t);
extern DACS_ERR_T dacspi_waitQe_alloc(dacs_wid_t wid, de_id_t de, 
                                      dacs_process_id_t pid, 
                                      dacsi_wid_elem_t **we, 
                                      dacs_tag_t *hwtag);
extern dacs_addr_64_t dacspi_waitQe_free (dacsi_wait_queue_t *waitQ,
                                          dacsi_wid_elem_t *waitQe); 

extern uint32_t dacspi_jobid;

// Today we only support the use of the first PIDS slot in the topology
// structure.  This define is used for this single index, but it also provides
// an easy way to locate all the code that would need to be adjusted if
// multiple pids/tids were supported.
#define ONLYPID 0

//  Initialize the JOB Q mutex        
#define DACSPI_JOBQ_LOCK_INIT(X)

// Locks the JOB Q mutex
#define DACSPI_JOBQ_LOCK(__Qidx, __Qtype)                                   \
    do {                                                                    \
        _mutex_lock ((mutex_ea_t)dacspi_get_jobQ((__Qidx), (__Qtype)));     \
    } while (0)

// Locks the JOB Q mutex
#define DACSPI_JOBQ_UNLOCK(__Qidx, __Qtype)                                 \
    do {                                                                    \
        _mutex_unlock ((mutex_ea_t)dacspi_get_jobQ((__Qidx), (__Qtype)));   \
    } while (0)

#define dacspi_get_jobQ(__Qidx, __Qtype) (&dacspi_jobQ[(__Qidx)][(__Qtype)])

#define DACSI_WE_ALLOC()                                                    \
        (malloc(sizeof(dacsi_wid_elem_t) + sizeof(dacsi_waitQ_t)))
#define DACSI_WE_FREE(__widQe) (free(__widQe))

#define DACSI_WAITQE_ALLOC() (malloc(sizeof(dacsi_waitQe_t)))
#define DACSI_WAITQE_FREE(__waitQe) (free(__waitQe))

#define DACSI_TAG_RESERVE() dacspi_tag_reserve()
#define DACSI_TAG_RELEASE(__tag) dacspi_tag_release(__tag)

extern dacsi_cb_t    dacs_rcb[1+NUM_SPES_PER_BLADE];
extern unsigned int  dacs_spu_finished[];



//  -----   topology  --------------------------------------------------

//  internal globals
extern dacs_topology_t dacs_tcb;
extern uint32_t dacs_topo_lock;

#define DACSPI_TOPO_LOCK()   _mutex_lock((mutex_ea_t)&dacs_topo_lock)
#define DACSPI_TOPO_UNLOCK() _mutex_unlock((mutex_ea_t)&dacs_topo_lock)

#define DACSPI_TOPO_CHILD(i) (&dacs_tcb.children[(i)])

extern  DACSI_TOPO_STS_T dacspi_check_reservation(  de_id_t de );
extern  int dacspi_check_children_reservations( dacs_topology_t *de_topo,
                                                DACSI_TOPO_STS_T checkval );
extern  int dacspi_set_children_reservations(    dacs_topology_t *de_topo,
                                                DACSI_TOPO_STS_T setval );

//  -----   test/wait   ------------------------------------------------
extern DACS_ERR_T dacs_ppu_test (dacsi_wait_queue_t *, dacsi_wid_elem_t *);
extern DACS_ERR_T dacs_ppu_wait (dacsi_wait_queue_t *, dacsi_wid_elem_t *);


//  =====   PPU "CONSTRUCTOR" CALLS

extern  DACS_ERR_T dacspi_topology_init(    void *argvp, void *envp );
extern  DACS_ERR_T dacspi_process_init(     void *argvp, void *envp );
extern  DACS_ERR_T dacspi_groups_init(      void *argvp, void *envp );
extern  DACS_ERR_T dacspi_tags_init(        void *argvp, void *envp );
extern  DACS_ERR_T dacspi_wids_init(        void *argvp, void *envp );
extern  DACS_ERR_T dacspi_mailbox_init(     void *argvp, void *envp );
extern  DACS_ERR_T dacspi_send_recv_init(   void *argvp, void *envp );
extern  DACS_ERR_T dacspi_data_sync_init(   void *argvp, void *envp );
extern  DACS_ERR_T dacspi_proc_sync_init(   void *argvp, void *envp );
extern  DACS_ERR_T dacspi_dma_init(         void *argvp, void *envp );
extern  DACS_ERR_T dacspi_shared_obj_init(  void *argvp, void *envp );

//  =====   PPU "DESTRUCTOR" CALLS

extern  DACS_ERR_T dacspi_topology_exit(    void );
extern  DACS_ERR_T dacspi_process_exit(     void );
extern  DACS_ERR_T dacspi_groups_exit(      void );
extern  DACS_ERR_T dacspi_tags_exit(        void );
extern  DACS_ERR_T dacspi_wids_exit(        void );
extern  DACS_ERR_T dacspi_mailbox_exit(     void );
extern  DACS_ERR_T dacspi_send_recv_exit(   void );
extern  DACS_ERR_T dacspi_data_sync_exit(   void );
extern  DACS_ERR_T dacspi_proc_sync_exit(   void );
extern  DACS_ERR_T dacspi_dma_exit(         void );
extern  DACS_ERR_T dacspi_shared_obj_exit(  void );

extern uint32_t dacspi_dma_lock;

#ifndef DACS_USE_LIBSPE2
#define __spe_mfcio_get(topo, ls, ea, len, tag) \
            _spe_mfc_dma((void *)(topo)->mfc_area, ls, (uint64_t)ea, len, tag, MFC_GET_CMD)
#define __spe_mfcio_getf(topo, ls, ea, len, tag) \
            _spe_mfc_dma((void *)(topo)->mfc_area, ls, (uint64_t)ea, len, tag, MFC_GETF_CMD)
#define __spe_mfcio_getb(topo, ls, ea, len, tag) \
            _spe_mfc_dma((void *)(topo)->mfc_area, ls, (uint64_t)ea, len, tag, MFC_GETB_CMD)
#define __spe_mfcio_put(topo, ls, ea, len, tag) \
            _spe_mfc_dma((void *)(topo)->mfc_area, ls, (uint64_t)ea, len, tag, MFC_PUT_CMD)
#define __spe_mfcio_putf(topo, ls, ea, len, tag) \
            _spe_mfc_dma((void *)(topo)->mfc_area, ls, (uint64_t)ea, len, tag, MFC_PUTF_CMD)
#define __spe_mfcio_putb(topo, ls, ea, len, tag) \
            _spe_mfc_dma((void *)(topo)->mfc_area, ls, (uint64_t)ea, len, tag, MFC_PUTB_CMD)
#else
#define __spe_mfcio_get(topo, ls, ea, len, tag) \
            spe_mfcio_get((spe_context_ptr_t)((topo)->ctx), \
                            ls, ea, len, tag, 0, 0)
#define __spe_mfcio_getf(topo, ls, ea, len, tag) \
            spe_mfcio_getf((spe_context_ptr_t)((topo)->ctx), \
                            ls, ea, len, tag, 0, 0)
#define __spe_mfcio_getb(topo, ls, ea, len, tag) \
            spe_mfcio_getb((spe_context_ptr_t)((topo)->ctx), \
                            ls, ea, len, tag, 0, 0)
#define __spe_mfcio_put(topo, ls, ea, len, tag) \
            spe_mfcio_put((spe_context_ptr_t)((topo)->ctx),  \
                            ls, ea, len, tag, 0, 0)
#define __spe_mfcio_putf(topo, ls, ea, len, tag) \
            spe_mfcio_putf((spe_context_ptr_t)((topo)->ctx), \
                            ls, ea, len, tag, 0, 0)
#define __spe_mfcio_putb(topo, ls, ea, len, tag) \
            spe_mfcio_putb((spe_context_ptr_t)((topo)->ctx), \
                            ls, ea, len, tag, 0, 0)
#endif

#define DACSPI_DMA(__op, __topo, __ls, __ea, __len, __tag)                   \
    do {                                                                    \
        int __m = ~0xf, __err = 0;                                          \
        unsigned int __sz;                                                  \
                                                                            \
        _mutex_lock((mutex_ea_t)(&(__topo)->dma_lock));                     \
        while ((__len) > 0 && !__err) {                                     \
            __sz = (((__len) & __m) >= MAX_DMA_SIZE)                        \
                        ? MAX_DMA_SIZE :(__len) & __m;                      \
            if (__sz) {                                                     \
                __err = __spe_mfcio_##__op((__topo), __ls,                  \
                                       (void*)__ea, __sz, (__tag));         \
                (__len) -= __sz;                                            \
                __ls += __sz;                                               \
                __ea += __sz;                                               \
            }                                                               \
            __m >>= !((__len) & __m);                                       \
        }                                                                   \
        _mutex_unlock((mutex_ea_t)(&(__topo)->dma_lock));                   \
                                                                            \
        return ((__err == -1) ? DACS_ERR_SYSTEM : DACS_SUCCESS);            \
    } while(0)

static inline DACS_ERR_T 
dacspi_dma_put(dacs_topology_t *topo, uint32_t ls, void *ea, unsigned int len, 
               dacs_tag_t tag)
{
    DACSPI_DMA(get, topo, ls, ea, len, tag);
    // The above macro returns, so we'll never get here
}

static inline DACS_ERR_T 
dacspi_dma_putf(dacs_topology_t *topo, uint32_t ls, void *ea, unsigned int len, 
                dacs_tag_t tag)
{
    DACSPI_DMA(getf, topo, ls, ea, len, tag);
    // The above macro returns, so we'll never get here
}

static inline DACS_ERR_T 
dacspi_dma_putb(dacs_topology_t *topo, uint32_t ls, void *ea, unsigned int len, 
                dacs_tag_t tag)
{
    DACSPI_DMA(getb, topo, ls, ea, len, tag);
    // The above macro returns, so we'll never get here
}

static inline DACS_ERR_T 
dacspi_dma_get(dacs_topology_t *topo, uint32_t ls, void *ea, unsigned int len, 
               dacs_tag_t tag)
{
    DACSPI_DMA(put, topo, ls, ea, len, tag);
    // The above macro returns, so we'll never get here
}

static inline DACS_ERR_T 
dacspi_dma_getf(dacs_topology_t *topo, uint32_t ls, void *ea, unsigned int len, 
                dacs_tag_t tag)
{
    DACSPI_DMA(putf, topo, ls, ea, len, tag);
    // The above macro returns, so we'll never get here
}

static inline DACS_ERR_T 
dacspi_dma_getb(dacs_topology_t *topo, uint32_t ls, void *ea, unsigned int len, 
                dacs_tag_t tag)
{
    DACSPI_DMA(putb, topo, ls, ea, len, tag);
    // The above macro returns, so we'll never get here
}

static inline DACS_ERR_T 
dacspi_dma_wait(dacs_topology_t *topo, dacs_tag_t tag) 
{
#ifndef DACS_USE_LIBSPE2
    int rc;
    _spe_mfc_write_tag_mask((void *)topo->mfc_area, (1<<tag));
    while (!_spe_mfc_read_tag_status_immediate((void *)topo->mfc_area));
    return DACS_SUCCESS;
#else
    unsigned int ret_mask;
    int rc = spe_mfcio_tag_status_read((spe_context_ptr_t)topo->ctx, 1 << tag, 
                                       SPE_TAG_ALL, &ret_mask);
    if (rc == -1)
        fprintf(stderr, "wait error: %d\n", errno);

    return ((rc == -1) ?  DACS_ERR_SYSTEM : 
            ((1 << tag) & rc) ? DACS_WID_READY : DACS_ERR_SYSTEM);
#endif
}

static inline DACS_ERR_T 
dacspi_dma_test(dacs_topology_t *topo, dacs_tag_t tag) 
{
#ifndef DACS_USE_LIBSPE2
    if (_spe_mfc_read_tag_status_immediate((void *)topo->mfc_area) & (1<<tag))
        return DACS_WID_READY;
    else
        return DACS_WID_BUSY;
#else
    unsigned int ret_mask;
    int rc = spe_mfcio_tag_status_read((spe_context_ptr_t)topo->ctx, 1 << tag, 
                                       SPE_TAG_IMMEDIATE, &ret_mask);
    return ((rc == -1) ?  DACS_ERR_SYSTEM : 
            ((1 << tag) & rc) ? DACS_WID_READY : DACS_WID_BUSY);
#endif
}

static inline int 
dacspi_ppu_mbox_status(dacs_context_t ctx)
{
    return spe_out_mbox_status((spe_context_ptr_t)ctx);
}

static inline int 
dacspi_spu_mbox_status(dacs_context_t ctx)
{
    return spe_in_mbox_status((spe_context_ptr_t)ctx);
}

static inline DACS_ERR_T 
dacspi_mbox_read(dacs_context_t ctx, unsigned int *data, int count)
{
    int ret;

    ret = spe_out_mbox_read((spe_context_ptr_t)ctx, data, count);

    if (ret == -1)
        return errno;
    else if (ret != count)
        return DACS_ERR_INTERNAL;
    else
        return DACS_SUCCESS;
}

static inline DACS_ERR_T 
dacspi_mbox_write(dacs_context_t ctx, unsigned int *data, int count, 
                  unsigned int behavior)
{
    int ret;

    ret = spe_in_mbox_write((spe_context_ptr_t)ctx, data, count, behavior);

    if (ret == -1)
        return errno;
    else if (ret != count)
        return DACS_ERR_INTERNAL;
    else
        return DACS_SUCCESS;
}

typedef union {
    unsigned long long ull;
    unsigned int ui[2];
} addr64;


#endif  //  _DACS_PPE_INTERNAL_H_
