/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
static char *ident = "@(#) internal-src/lib/dacs/cell/ppu/dacs_ppu_data_sync.c v1.35 - 9/24/07 13:33:13 @(#)";

// Support for data synchronization on PPE

#include <dacs_ppe_internal.h>
#include <dacsi_q.h>
#include <dacsi_wids.h>
#include <dacs_debug.h>
#include <dacs.h>

//  -----   local prototypes    -----
DACS_ERR_T dacsi_ppu_test (dacsi_wait_queue_t *, dacsi_wid_elem_t *, int);

//  -----   Globals -----

DACS_ERR_T  
dacs_ppu_test (dacsi_wait_queue_t *waitQ, dacsi_wid_elem_t *we)
{
    return dacsi_ppu_test(waitQ, we, 0);
}

DACS_ERR_T 
dacs_ppu_wait (dacsi_wait_queue_t *waitQ, dacsi_wid_elem_t *we)
{
    return dacsi_ppu_test(waitQ, we, 1);
}

DACS_ERR_T
dacsi_ppu_test (dacsi_wait_queue_t *waitQ, dacsi_wid_elem_t *we, int waitok)
{
    dacs_topology_t *topo = NULL;
    DACS_ERR_T err = DACS_SUCCESS;
    DACSI_JOB_TYPE_T type;
    volatile dacsi_jobQe_t *job;
#if NEED_DEBUG
    int already_printed = 0;
#endif

    DPRINT("Starting wait on element %p\n", we);

#ifdef DACS_ERROR_CHECKING
    if (!dacsi_is_initialized()) { 
        return DACS_ERR_NOT_INITIALIZED; 
    }
#endif

    // Fetch the topology struct for the source de
    if ((topo = dacsi_get_topo(we->de, &err)) == NULL) {
        return  DACS_ERR_INVALID_DE;
    }

    while ((type = dacsi_waitQe_get(we, (void*)&job))) {

        DPRINT("Got job type %d\n", type);

        // PUTs and GETs do it the easy way because there is no hand-shake
        if (type != DACSI_JOB_TYPE_SEND && type != DACSI_JOB_TYPE_RECV) {
            // Wait on the hwtag
            if (waitok)
                err = dacspi_dma_wait(topo, we->hwtag); 
            else {
                err = dacspi_dma_test(topo, we->hwtag); 
                if (err == DACS_WID_BUSY)
                    return err;
            }
        }

        // SEND and RECEIVE ... do it the hard way!
        else {
            volatile dacsi_jobQ_t *jobQ;
            DACSI_JOB_ST_T state = 0;
            DACSI_JOBQ_TYPE_T Qtype;
            unsigned int Qidx;
            int next = 0;

            DPRINT("Got job id %d\n", job->jobid);

#ifdef DACS_ERROR_CHECKING
            if (job->state == DACSI_JOB_ST_INVALID ||
                job->state == DACSI_JOB_ST_FREE)
            {
                DPRINT("Invalid job attached to wait Q element %p.\n", we);
                return  DACS_ERR_INVALID_WID;
            }
#endif
            Qtype = job->Qtype;
            Qidx = job->Qidx;

            do {
                DACSPI_JOBQ_LOCK(Qidx, Qtype);
                state = job->state;
                DACSPI_JOBQ_UNLOCK(Qidx, Qtype);

                switch (state) {
                    case DACSI_JOB_ST_FINISHED:
                    case DACSI_JOB_ST_COMPLETED:
                    {
                        DACSPI_JOBQ_LOCK(Qidx, Qtype);

                        err = job->error;
                        state = job->state;

                        // The first participant to see the job complete will 
                        // set the state to COMPLETED as a signal to the 
                        // second to remove the job.  If COMPLETED then we 
                        // can free the job.
                        // Otherwise, we are the first in so clear the 
                        // initiator and get out.
                        if (state == DACSI_JOB_ST_COMPLETED) {
                            DPRINT("Job ID %d COMPLETED with %s\n", 
                                    job->jobid, dacs_strerror(err));
                            // Pull this job off the Q
                            dacsi_jobQe_free((dacsi_jobQe_t *)job);
                        }
                        else {
                            DPRINT("Job ID %d FINISHED with %s\n", 
                                    job->jobid, dacs_strerror(err));
                            ASSERT(job->state == DACSI_JOB_ST_FINISHED);
                            job->state = DACSI_JOB_ST_COMPLETED;
                        }

                        DACSPI_JOBQ_UNLOCK(Qidx, Qtype);

                        // Move on to the next job under this wid
                        next = 1;
                        break;
                    }

                    case DACSI_JOB_ST_PENDING:
                    {
                        // The job has been started, meaning DMA is underway.  
#ifdef NEED_DEBUG
                        if (!already_printed) {
                            DPRINT("Job ID %d is pending.\n", job->jobid);
                            already_printed = 1;
                        }
#endif
                        if (waitok) {
                            if (job->initiator == dacspi_local_de_id) {
                                // This is the initiator, so go ahead and do a 
                                // hardware wait.
                                DPRINT("Initiator of jobID %d wait on tag %d\n",
                                       job->jobid, we->hwtag);

                                err = dacspi_dma_wait(topo, we->hwtag);

                                DPRINT("DMA wait completed, err %s\n", 
                                        dacs_strerror(err));
                            }
                        }
                        else {
                            // The job is still pending, so we don't need to
                            // test any further, just return with busy.
                            return DACS_WID_BUSY;
                        }

                        break;
                    }

                    default:
                    {
                        // If we get here then we have an unexpected state 
                        // meaning something went drastically wrong.
                        err = DACS_ERR_INTERNAL;
                        DPRINT("Job ID %d has an unexpected state %d\n", 
                                job->jobid, state);
                        break;
                    }
                }
            } while (waitok && !err && !next); 
        }
        // We are done with this job and it's associated wait Q entry.  We can
        // release this entry and move onto the next one if one exists.
        dacsi_waitQe_free(we);
    }

    // We are done servicing all the jobs under this wid element, so we can
    // release it and the wait Q header.
    dacsi_we_free(waitQ, we);

    DPRINT("Exiting with %d\n", err);

    return err;
}


//  =====   CONSTRUCTOR and DESTRUCTOR

//
//
//
DACS_ERR_T  dacspi_data_sync_init( void *argvp, void *envp )
{
    dacsi_wids_init();
    return  DACS_SUCCESS;
}


//
//
//
DACS_ERR_T  dacspi_data_sync_exit( void )
{
    dacsi_wait_queue_t *q;
    dacsi_wid_elem_t *elem;
    dacs_addr_64_t *req;
    int wid;

    // Walk through all of our wids and free any waitQe and wid_elem
    // that have been left behind.
    // We have the means for freeing the jobQ stuff here too, but since it is
    // static storage that gets cleared on init, it is not necessary.
    for (wid = 0; wid < DACSI_WID_MAX; wid++) {
        DACSI_LOCK_WID_QUEUE(wid);
        q = &dacsi_waitq[wid];
        elem = q->head;

        while (elem) {
            while (dacsi_waitQe_get(elem, (void *)&req)) {
                dacsi_waitQe_free(elem);
            }
            dacsi_we_free(q, elem);
            elem = q->head;
        }
        DACSI_UNLOCK_WID_QUEUE(wid);
    }
    dacsi_wids_destroy();
    return  DACS_SUCCESS;
}
