/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

static char *ident = "@(#) internal-src/lib/dacs/cell/ppu/dacs_ppu_runtime.c v1.43 - 9/26/07 11:53:28 @(#)";

//
// Runtime initialization for PPE
//

#include    <unistd.h>
#include    <dacs_hybrid.h>
#include    <dacs_ppe_internal.h>
#include    <dacs_hybrid_topology.h>
#include    <dacs_debug.h>

//
//  add all the detailed function descriptions to the PPE LLD
//  Note that REPEAT_BRIEF is on in the Doxyfile, so the @brief descriptions
//  in the .h file will be prepended to these detailed descriptions - no need
//  to repeat them here.
//

//  -----   local prototypes    -----
static  void        set_spe_cb_entry(  uint32_t         index,
                                       de_id_t          de_id,
                                       DACS_DE_TYPE_T   type
                                    );


//  -----   Globals ----------------------------------------------------
// The local de_id of the PPE
de_id_t     dacspi_local_de_id = NULL_DE_ID;

//
// flag showing that dacs_runtime_init() has been run
uint32_t   dacsi_runtime_initialized    =   0;      // init to FALSE

//
// Runtime Control Block array for PPE and its' children.
// Index 0 is reserved for PPE, indices 1-N are reserved for
// the SPE's.
// Since the PPE can be configured as either DACS_DE_BLADE (16 spe's),
// or DACS_DE_CBE (with 8 spe's), we will allocate the maximum size
// here.
// This block should only be used for \b static information
// since it will be downloaded \b once during dacs_runtime_init.
//
dacsi_cb_t   dacs_rcb[1+NUM_SPES_PER_BLADE] __attribute__ ((aligned (128)));

// Array of SPE finished template LS addresses
unsigned int dacs_spu_finished[1+NUM_SPES_PER_BLADE] = {0};

//
// PPU Runtime Initialization
// This initializes the DaCS "interface" with the
// children DE's
// dacs_hybrid_runtime_init() should have run already and found out who and
// what we are.
//
// We check here to see what our DE type is.
// If it's DACS_DE_CBE, we own 8 spe's, if it's DACS_DE_BLADE, we own all 16.
// (8 and 16 are the values of NUM_SPES_PER_CBE/NUM_SPES_PER_BLADE of course)
//
DACS_ERR_T  dacs_ppu_runtime_init(  void *argvp,
                                void *envp,
                                de_id_t de_to_pass
                                )
{
    int   i                 =   0;
    DACS_DE_TYPE_T  my_type =   DACS_DE_INVALID;
    DACS_ERR_T  err_code    =   DACS_SUCCESS;


    if(dacsi_runtime_initialized)
    {
        DPRINT("dacs_ppu_runtime_init has already been callen on the PPE!!!\n");
        return DACS_ERR_INITIALIZED;
    }

    /*  critical section    */

    DPRINT( "hybrid layer passed us a de_id=0x%x\n",    de_to_pass );
    dacspi_local_de_id  =   de_to_pass;      // store local copy

    //  parse our de to determine if we're a blade or a CBE
    if ((dacspi_local_de_id & 0x0000FF00) != 0)
    {
        my_type = DACS_DE_CBE;
    }
    else if((dacspi_local_de_id & 0x00FF0000) != 0 )
    {
        my_type = DACS_DE_CELL_BLADE;
    }

    if ( err_code != DACS_SUCCESS )
    {
        DPRINT( "dacs_hybrid_topology_get_type returned 0x%x, rc=%d\n",
                my_type, err_code );
        // don't update de if failure
        return err_code;
    }

    DPRINT( "fill in dacs_rcb[0]\n" );
    //  Then fill in the PPE control block entry
        set_spe_cb_entry(0, dacspi_local_de_id, my_type);

    DPRINT( "fill in dacs_rcb[1-N]\n" );
    //
    //  Fill in SPE control blocks
    //
    for (i = 1; i < (NUM_SPES_PER_BLADE)+1; i++ )
    {
        set_spe_cb_entry(i, (dacspi_local_de_id+i), DACS_DE_SPE);
    }

    //
    //  initialize all ppe modules
    //
    DPRINT( "run topology init\n");
    err_code = dacspi_topology_init(argvp,envp);

    //  set up process control
    if ( err_code == DACS_SUCCESS ) {
        DPRINT( "run process init\n");
        err_code = dacspi_process_init( argvp,envp );
    }

    //  set up dacs groups structures
    if ( err_code == DACS_SUCCESS ) {
        DPRINT( "run groups init\n");
        err_code = dacspi_groups_init( argvp,envp );
    }       
    
    //  set up dacs shared object structures
    if ( err_code == DACS_SUCCESS ) {
        DPRINT( "run shared_obj init\n");
        err_code = dacspi_shared_obj_init( argvp,envp );
    }

    //  set up dacs tags structures
    if ( err_code == DACS_SUCCESS ) {
        DPRINT( "run tags init\n");
        err_code = dacspi_tags_init( argvp,envp );
    }

    //  set up mailbox queues
    if ( err_code == DACS_SUCCESS ) {
        DPRINT( "run mailbox init\n");
        err_code = dacspi_mailbox_init( argvp,envp );
    }

    //  set up messaging
    if ( err_code == DACS_SUCCESS ) {
        DPRINT( "run send_recv init\n");
        err_code = dacspi_send_recv_init( argvp,envp );
    }

    //  set up data sync
    if ( err_code == DACS_SUCCESS ) {
        DPRINT( "run data_sync init\n");
        err_code = dacspi_data_sync_init( argvp,envp );
    }

    //  set up process sync
    if ( err_code == DACS_SUCCESS ) {
        DPRINT( "run proc_sync init\n");
        err_code = dacspi_proc_sync_init( argvp,envp );
    }

    //  set up dma / windowing
    if ( err_code == DACS_SUCCESS ) {
        DPRINT( "run dma init\n");
        err_code = dacspi_dma_init( argvp,envp );
    }

    // set runtime flag to true
    if ( err_code == DACS_SUCCESS )
        dacsi_runtime_initialized = 1;
     
    // If we had an error, we need to clean things up
    if (err_code != DACS_SUCCESS)
        dacs_runtime_exit();

    /*  end critical section */

    return err_code;
}

//
//  Wrapper for PPU runtime exit.
//
DACS_ERR_T  dacs_ppu_runtime_exit( void )
{
    DACS_ERR_T  err_code  =   DACS_SUCCESS;

    /* critical section */

    // What ?!?!? We're not even initialized.  Return an error.
    if (!dacsi_runtime_initialized)
        return DACS_ERR_NOT_INITIALIZED;

    //  - terminate all ppe modules
    err_code = dacspi_dma_exit(   );          //  set up dma / windowing
    if ( err_code != DACS_SUCCESS ) {  return ( err_code );    }
    err_code = dacspi_proc_sync_exit(   );    //  set up process sync
    if ( err_code != DACS_SUCCESS ) {  return ( err_code );    }
    err_code = dacspi_data_sync_exit(   );    //  set up data sync
    if ( err_code != DACS_SUCCESS ) {  return ( err_code );    }
    err_code = dacspi_send_recv_exit(   );    //  set up messaging
    if ( err_code != DACS_SUCCESS ) {  return ( err_code );    }
    err_code = dacspi_mailbox_exit(   );      //  set up mailbox queues
    if ( err_code != DACS_SUCCESS ) {  return ( err_code );    }
    err_code = dacspi_shared_obj_exit(   );   // close down shared_objs
    if ( err_code != DACS_SUCCESS ) {  return ( err_code );    }
    err_code = dacspi_tags_exit(   );         //  set up dacs tags structures
    if ( err_code != DACS_SUCCESS ) {  return ( err_code );    }
    err_code = dacspi_groups_exit(   );       //  set up dacs groups structures
    if ( err_code != DACS_SUCCESS ) {  return ( err_code );    }
    err_code = dacspi_process_exit(   );      //  set up process control
    if ( err_code != DACS_SUCCESS ) {  return ( err_code );    }
    err_code = dacspi_topology_exit( );
    if ( err_code != DACS_SUCCESS ) {   return ( err_code );    }

    // set runtime flag back to false 
    // "Unit Test exited abnormally (status: b)!")
    dacsi_runtime_initialized    =   0;      

    /* end critical section */

    return  DACS_SUCCESS;
}


//  ====================    PRIVATE ROUTINES    ========================


//  ----- PPE->SPE  runtime --------------------------------------------

void 
dacsi_dump_cb( const char *offset, dacsi_cb_t *cb )
{
    if ( cb == NULL )   { return;   }   // sanity check

    DPRINT( "%s dump cb @ %p:\n", offset, cb );
    DPRINT( "%s\t de_id         =   0x%x\n",    offset, (uint32_t)cb->de_id );
    DPRINT( "%s\t type          =   0x%x\n",    offset, (uint32_t)cb->type );
}

static void 
set_spe_cb_entry(uint32_t index, de_id_t de_id, DACS_DE_TYPE_T type)
{
    dacs_rcb[index].signature   =   DACS_CB_SIGNATURE;
    dacs_rcb[index].de_id       =   de_id;
    dacs_rcb[index].type        =   type;
}
