/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
static char *ident = "@(#) internal-src/lib/dacs/cell/ppu/dacs_ppu_send_recv.c v1.38 - 9/24/07 13:33:14 @(#)";

//
// Messaging API's for PPE
//

#include <unistd.h>
#include <stddef.h>
#include <stdlib.h>
#include <dacs_ppe_internal.h>
#include <dacs_ppu_send_recv.h>
#include <dacsi_q.h>
#include <dacs_debug.h>
#include <dacsi_wids.h>


// Starting job id for PPE.  Nonzero for debug
#define START_JOBID     10042

// counter for the jobid's submitted to jobs Q
uint32_t dacspi_jobid = START_JOBID;
static DACSI_JOB_ST_T finished = DACSI_JOB_ST_FINISHED;

//  -----   local prototypes    -----
static DACS_ERR_T common_input_check (de_id_t, dacs_process_id_t, void *, 
                                      uint32_t, dacs_wid_t, DACS_BYTE_SWAP_T);
static dacsi_jobQe_t *dacspi_create_job(DACSI_JOBQ_TYPE_T, uint32_t, uint32_t, 
                                        uint64_t, uint32_t);

//  -----   Globals -----
//  DACS message job Q
dacsi_jobQ_t dacspi_jobQ[1+NUM_SPES_PER_BLADE][DACSI_JOBQ_CNT] __attribute__ ((aligned(128))); 

// Array of SPU finished job templates
// This array contains the local store address of each SPU's finished job
// template.
extern unsigned int dacs_spu_finished[];

DACS_ERR_T 
dacspi_send_recv (void                *data,
                  uint32_t            len,
                  de_id_t             de,
                  dacs_process_id_t   pid,
                  uint32_t            stream,
                  dacs_wid_t          wid,
                  DACS_BYTE_SWAP_T    swap,
                  DACSI_JOBQ_TYPE_T   Qtype)
{
    DACS_ERR_T joberr = DACS_SUCCESS;
    DACS_ERR_T reterr = DACS_SUCCESS;
    dacs_topology_t *topo = NULL;
    dacsi_jobQ_t *jobQ;
    dacsi_jobQe_t *job;
    dacsi_wid_elem_t *we;
    dacsi_waitQe_t *waitQe;
    unsigned int hwtag = 0;

    DPRINT("Starting recieve from DE 0x%x PID 0x%x WID %d DATA %p LEN %d\n",
           (unsigned int)de, (unsigned int)pid, wid, data, len);

#ifdef DACS_ERROR_CHECKING
    // Now check for errors that we need to propagate to the other side.
    reterr = common_input_check(de, pid, data, len, wid, swap);

    if (reterr)
        return reterr;
#endif

    // Fetch the topology struct for the source de
    // Don't overwrite previously encountered errors
    topo = dacsi_get_topo(de, &reterr);
    if (!topo) {
        return DACS_ERR_INVALID_DE;
    }

#ifdef DACS_ERROR_CHECKING
    if (pid != topo->pids[ONLYPID].pid)
        return DACS_ERR_INVALID_PID;
#endif

    DACSPI_JOBQ_LOCK(topo->my_index, Qtype);

    // Get the source index.
    jobQ = dacspi_get_jobQ(topo->my_index, Qtype);

    //  Peek at the send Q to see if a recv is already there.
    //  We are under lock so we should be safe.
    job = dacsi_jobQe_get(jobQ, stream, (Qtype == DACSI_JOBQ_SEND) 
                            ? DACSI_JOB_TYPE_RECV : DACSI_JOB_TYPE_SEND);

    // A job was already posted but we got an error.  We need to fail the
    // queued job to let the sender knows we won't be participating.
    // The first guy in (job poster) is responsible for removing the job.
    if (!job) {
        // Find or allocate a wait queue element for this transaction.
        // Since we are not initiating things, we don't need a hwtag.
        reterr = dacsi_we_get(wid, de, pid, &we, &waitQe, NULL);

        // We must be the first in because there is no job, so we'll queue
        // the job.  The sender will be the initiator.
        // Don't Q the job if we failed to get a wait buffer.
        if (!reterr)
            job = dacspi_create_job(Qtype, stream, len, 
                                    (dacs_addr_64_t)data, topo->my_index);

        DPRINT("Created and queued job ID %d\n", job->jobid); 
    }
    else {
        unsigned int finished;
        job->initiator = dacspi_local_de_id;

        // A job already exists in the send Q, so we must have been the
        // second one in.
        DPRINT("Job already posted for job ID %d\n", job->jobid);

        // Check that the receive buffer is large enough.
        // If not, return an error.
        if ((Qtype == DACSI_JOBQ_RECV && len < job->len) ||
            (Qtype == DACSI_JOBQ_SEND && len > job->len)) {
            DPRINT("Receive buffer too small. %d < %d\n", len, job->len);
            joberr = DACS_ERR_BUF_OVERFLOW;
        }

        if ((job->ls & 0xf) != ((uint64_t)data & 0xf)) {
            DPRINT("Misaligned send/recv buffers. ls 0x%x ms %p\n",
                    job->ls, data);
            joberr = DACS_ERR_NOT_ALIGNED;
        }

        // We were 2nd in so we'll initiate the DMA.
        // First, fetch the address of the target's finished template, so we
        // can issue the finished DMA immediately after the message DMA.  
        // TODO: Change the template array to point directly at the state
        // field.
        finished = dacs_spu_finished[topo->my_index] + 
                       offsetof(dacsi_jobQe_t, state);

        // Find or allocate a wait queue element for this transaction.
        // We are initiating the DMA so we need to pass a pointer for
        // returning a hwtag.
        reterr = dacsi_we_get(wid, de, pid, &we, &waitQe, &hwtag);

        // Initiate DMA on given job.
        // NOTE: The contents of the job are unpredicatable after the 
        //       fenced get, since it clears everything and marks things
        //        finished.
        if (!reterr && !joberr) {
            if (Qtype == DACSI_JOBQ_SEND)
                reterr = dacspi_dma_put(topo, job->ls, data, job->len, hwtag);
            else
                reterr = dacspi_dma_get(topo, job->ls, data, job->len, hwtag);

            // If we encountered an error to be returned, then we don't want to
            // mark the job finished
            joberr = dacspi_dma_getf(topo, finished, &job->state, 
                                  sizeof(job->state), hwtag);
        }

        DPRINT("Started DMA for job ID %d on hwtag %d err = %s\n", 
                job->jobid, hwtag, dacs_strerror(joberr));

        // Check if we experienced an error to pass on to the recipient
        // If so, mark the job failed with the error reason.
        if (joberr) {
            job->state = DACSI_JOB_ST_FINISHED;
            job->error = joberr;
            DPRINT("Marking existing job as failed. error = %d\n", joberr);
        }
    }

    DACSPI_JOBQ_UNLOCK(topo->my_index, Qtype);

    // If we did not experience an error that Q the job for wait to find.
    if (!reterr) {
        reterr = dacsi_waitQe_enq(wid, we, waitQe, (Qtype == DACSI_JOBQ_SEND) 
                                  ? DACSI_JOB_TYPE_SEND : DACSI_JOB_TYPE_RECV, 
                                  (dacs_addr_64_t)job);

        DPRINT("Added job ID %d to wait Q element %p\n", job->jobid, we);
    }

    // If we failed for a reason other than failing to get a wait buffer then
    // free the buffer.
    else if (reterr != DACS_ERR_NO_RESOURCE) {
        DACSI_LOCK_WID_QUEUE(wid);
        dacsi_we_free(NULL, we);
        DACSI_UNLOCK_WID_QUEUE(wid);
    }

    return reterr;
}

DACS_ERR_T 
dacs_ppu_send (void               *src_data,
               uint32_t            len,
               de_id_t             dst_de,
               dacs_process_id_t   dst_pid,
               uint32_t            stream,
               dacs_wid_t          wid,
               DACS_BYTE_SWAP_T    swap)
{
    return dacspi_send_recv(src_data, len, dst_de, dst_pid, stream, wid, swap,
                            DACSI_JOBQ_SEND);
}

DACS_ERR_T 
dacs_ppu_recv (void                *dst_data,
               uint32_t            len,
               de_id_t             src_de,
               dacs_process_id_t   src_pid,
               uint32_t            stream,
               dacs_wid_t          wid,
               DACS_BYTE_SWAP_T    swap)
{
    return dacspi_send_recv(dst_data, len, src_de, src_pid, stream, wid, swap,
                            DACSI_JOBQ_RECV);
}


//  =====   INTERNAL FUNCTIONS  ========================================
//
static DACS_ERR_T  
common_input_check(de_id_t de, dacs_process_id_t pid, void *data, 
                   uint32_t len, dacs_wid_t wid, DACS_BYTE_SWAP_T swap)
{
    // Check if DaCS is initialized.  
    if (!dacsi_is_initialized())  
        return DACS_ERR_NOT_INITIALIZED;

    // can't receive a message from yourself
    if (de == dacspi_local_de_id || de == DACS_DE_SELF || pid == DACS_PID_SELF)
        return DACS_ERR_INVALID_TARGET;

    // We must be using a valid wid 
    if (wid >= MAX_PPE_WIDS)
        return DACS_ERR_INVALID_WID;

    // The wid must have been reserved
    if (!DACSI_IS_WID_RESERVED(wid))
        return DACS_ERR_INVALID_WID;

    // check that we were passed a good buffer
    if (data == NULL)
        return DACS_ERR_INVALID_ADDR;

    if (swap)
        return DACS_ERR_NOT_SUPPORTED_YET;

    // The message buffer must be at least aligned on the first power-of-2 less
    // than the message len up to 16 bytes.
    int power = 0x8;

    if (len < 16)
        while (!(power & (len & 0xf))) power >>= 1;
    else
        power = 0x10;

    if ((uint64_t)data & (power-1)) {
        return DACS_ERR_NOT_ALIGNED;
    }

    return DACS_SUCCESS;
}

void
dacspi_jobQ_init(dacsi_jobQ_t *jobQ, int Qidx, int Qtype)
{
    int i;
    dacsi_jobQe_t *jobQe = jobQ->entries;

    for (i = 0; i < DACSI_JOBQE_CNT; i++) {
        jobQe[i].Q.off = 
            (unsigned int)((unsigned long)&jobQe[i] - (unsigned long)jobQ);
        jobQe[i].Q.next = (i < (DACSI_JOBQE_CNT-1)) 
            ? (unsigned int)((unsigned long)&jobQe[i+1] - (unsigned long)jobQ) 
            : (unsigned int)(offsetof(dacsi_Q_t, free));
        jobQe[i].Q.prev = (i > 0) 
            ? (unsigned int)((unsigned long)&jobQe[i-1] - (unsigned long)jobQ) 
            : (unsigned int)(offsetof(dacsi_Q_t, free));
        jobQe[i].Qtype = Qtype;
        jobQe[i].Qidx = Qidx;
    }

    jobQ->Q.free.next = jobQe[0].Q.off;
    jobQ->Q.free.prev = jobQe[DACSI_JOBQE_CNT-1].Q.off;
    jobQ->Q.free.off = offsetof(dacsi_Q_t, free);
    jobQ->Q.active.next = jobQ->Q.active.prev = offsetof(dacsi_Q_t, active);
    jobQ->Q.active.off = offsetof(dacsi_Q_t, active);
}

dacsi_jobQe_t *
dacspi_create_job(DACSI_JOBQ_TYPE_T Qtype, uint32_t stream, uint32_t len, 
                  uint64_t buf, uint32_t Qidx)
{
    dacsi_jobQe_t *job;

    job = dacsi_jobQe_alloc(&dacspi_jobQ[Qidx][Qtype]);
    if (!job)
        return NULL;

    job->type      = (Qtype == DACSI_JOBQ_SEND)
                        ? DACSI_JOB_TYPE_SEND : DACSI_JOB_TYPE_RECV;
    job->state     = DACSI_JOB_ST_PENDING;
    job->error     = DACS_SUCCESS;
    job->swap      = (uint8_t)0;
    job->len       = len;
    job->ms        = (uint64_t)buf;
    job->stream    = stream;
    job->initiator = 0;
    job->jobid     = dacspi_jobid++;
    job->Qtype     = Qtype;
    job->Qidx      = Qidx;

    return dacsi_jobQe_put(job);
}

//  =====   CONSTRUCTOR and DESTRUCTOR

//  --------------------------------------------------------------------
// "Constructor" for messaging functions
// --------------------------------------------------------------------
DACS_ERR_T dacspi_send_recv_init(void *argvp, void *envp)
{
    unsigned int q, speidx;

    dacspi_jobid = START_JOBID;      // reset jobid counter

    for (speidx = 0; speidx < 1+NUM_SPES_PER_BLADE; speidx++) {
        for (q = 0; q < DACSI_JOBQ_CNT; q++) {
            dacspi_jobQ_init(&dacspi_jobQ[speidx][q], speidx, q);
            dacs_rcb[speidx].jobQ[q] = (dacs_addr_64_t)&dacspi_jobQ[speidx][q];
        }
    }

    return DACS_SUCCESS;
}


// --------------------------------------------------------------------
// "Destructor" for messaging functions
// --------------------------------------------------------------------
DACS_ERR_T  dacspi_send_recv_exit( void )
{
    return  DACS_SUCCESS;
}
