/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

static char *ident = "@(#) internal-src/lib/dacs/cell/ppu/dacs_ppu_shared_obj.c v1.12 - 9/24/07 13:18:11 @(#)";

#include <dacs.h>
#include <dacs_ppe_internal.h>
#include <dacs_common.h>
#include <dacs.h>
#include <dacs_ppu_data_sync.h>
#include <dacs_ppu_send_recv.h>

#include <errno.h>

#define _XOPEN_SOURCE 600     // required for for posix memalign 
#include <stdlib.h>           // for posix_memalign

#include <dacs_debug.h>

#define dacspi_local_pid 0

//  -----   local prototypes    -----

// dacspi_ppu_shared_obj_init is used to malloc and initialize all of the 
// parts of the shared_obj that are not unique (parts not specific to 
// shared_mem, mutex, barrier).  The caller then initializes their 
// object-specific version of the structs
DACS_ERR_T  dacspi_ppu_shared_obj_create( dacsi_shared_obj_t **obj )
{
    return DACS_SUCCESS;
}

// dacspi_ppu_share_obj uses a similiar mechanism to send/recv, and 
// also is responsible for allocating and deallocating a wid to wait on.
DACS_ERR_T dacspi_ppu_share_obj( de_id_t              dst_de,
                                 dacs_process_id_t    dst_pid,
                                 dacsi_shared_obj_t  *obj )
{
    uint32_t                 dst_index;
    dacs_topology_t         *topo;
    DACS_ERR_T               err;
    dacs_wid_t               wid;
    uint32_t                 stream = 0;
    ALIGNED_DATA(send_msg, dacsi_shared_obj_t *, 16);
    ALIGNED_DATA(recv_msg, DACSI_SHARED_OBJ_NAME_T, 16);

    topo = dacsi_get_topo(dst_de, &err);
    if(!topo) return DACS_ERR_INVALID_DE;
    dst_index   =   topo->my_index;

    // We directly use the internal version of the send function, but instead of
    // using the send/receive queues(mcbs), we use the internal _sendmcbs
    DPRINT("Sharing obj with SPU %d\n", dst_index);
    DPRINT("        obj->owner_de  = 0x%x\n",   obj->owner_de);
    DPRINT("        obj->owner_pid = 0x%x\n",   obj->owner_pid);
    DPRINT("        obj->tier      = 0x%x\n",   obj->tier);
    DPRINT("        obj->name      = 0x%x\n",   obj->name);
    if(obj->name == DACSI_MUTEX_NAME) {
    stream = DACSI_MUTEX_STREAM;
    } else if (obj->name == DACSI_REMOTE_MEM_NAME) {
    stream = DACSI_REMOTE_MEM_STREAM;
    } else if (obj->name == DACSI_GROUP_NAME) {
    stream = DACSI_GROUP_STREAM;
    } else {
       return DACS_ERR_INVALID_ATTR;
    }

    dacs_wid_reserve(&wid);

    // Send over a pointer to the object 
    *send_msg = obj;
    err = dacs_ppu_send (send_msg,                    //*src_data
                         sizeof(dacsi_shared_obj_t *), //len
                         dst_de,                       //dst_de
                         dst_pid,                      //dst_pid
                         stream,                       //stream
                         wid,                          //wid
                         0);                           //swap

    if (err != DACS_SUCCESS) 
        goto dacspi_ppu_share_obj_end;

    err = dacs_wait(wid);
    if (err != DACS_SUCCESS) 
        goto dacspi_ppu_share_obj_end;
 
    // Wait for the completion message from the other side
    err = dacs_ppu_recv (recv_msg,                    // *dst_data
                         sizeof(uint32_t),            // len
                         dst_de,                      // src_de
                         dst_pid,                     // src_pid
                         stream,                      // stream
                         wid,                         // wid
                         0);                          // swap

    if (err != DACS_SUCCESS) 
        goto dacspi_ppu_share_obj_end;

    err = dacs_wait(wid);
    if (err != DACS_SUCCESS) 
        goto dacspi_ppu_share_obj_end;

    // The acceptor should send back the ME key to complete the handshake 
    // successfully.  This is just something we use internally as a 
    // nontrivial identifier
    if(*recv_msg != obj->name) {
        err = DACS_ERR_INTERNAL;
        goto dacspi_ppu_share_obj_end;
    }

    // Atomically increment the refcount
    _atomic_modify((atomic_ea_t) &(obj->refcnt),1);

dacspi_ppu_share_obj_end:
    dacs_wid_release(&wid);
    return err;
}


DACS_ERR_T dacspi_ppu_shared_obj_accept(de_id_t src_de,
                                        dacs_process_id_t src_pid,
                                        dacsi_shared_obj_t *obj)
{
    return DACS_ERR_NOT_SUPPORTED_YET;
}

DACS_ERR_T dacspi_ppu_shared_obj_release( dacsi_shared_obj_t * obj)
{
    return DACS_ERR_NOT_SUPPORTED_YET;
}

//  =====   INTERNAL FUNCTIONS  ========================================

//  =====   CONSTRUCTOR and DESTRUCTOR

DACS_ERR_T  dacspi_shared_obj_init(void *argvp, void *envp)
{
    return  DACS_SUCCESS;
}

DACS_ERR_T  dacspi_shared_obj_exit(void)
{
    return  DACS_SUCCESS;
}

